import random
import matplotlib.pyplot as plt
import copy
DEBUG = False

class Pomdp:
    def __init__(self, b = 0.5):
        self.belief = b # 1 is certain state x1, 0 is certain x2
        # teda: p1 = belief, p2 = 1 - belief
        self.discount = 1 # 0 favors immidiate reward, 1 distant future rewads
        # rewards maps state, action to immidiate reward
        self.rewards = {('x1','u1') : -100,
                        ('x1','u2') : 100,
                        ('x1','u3') : -1,
                        ('x2','u1') : 100,
                        ('x2','u2') : -50,
                        ('x2','u3') : -1
                        }
        self.terminal = set(['u1', 'u2']) # set of terminal actions...
        self.actions = [['u1', 'u2', 'u3']]
        self.state_change_prop = {  'u1' : 0.0,
                                    'u2' : 0.0,
                                     'u3' : 0.8 
                                     }
        self.observations_prop = {('x1','z1') : 0.7, ('x2', 'x2'): 0.7 }
        #format: [action, [v1, v2]] - means list of linear functions: v1*p1 + v2*p2
        # self.value_vectors = [['u1',[-100, 100]], ['u2', [100, -50]], ['u3', [-1, -1]]  ]
        # for horizon 2, after pruning
        # self.value_vectors = [['u1',[-100, 100]], ['u2', [100, -50]], ['u3', [51, 42]]  ]

        #value fnc after executing u3, hardcoded, for testing pruposes
        # self.value_vectors = [['u2',[60, -60]], ['u1', [-20, 70]], ['u3', [52, 43]]  ]

    
    def pomdp(self,horizon, pruning=0):
        # returns value function over belief space?
        # returns set of vectors representing value function...
        value = [] # format : x1*p1 + x2*p2 represented as (x1,x2)
        value = [['u1',[-100, 100]], ['u2', [100, -50]] ]
        for i in range(horizon):
            val_z1 = copy.deepcopy(value)
            val_z2 = copy.deepcopy(value)
            for v in val_z1:
            # update belief according to observation
                #update for observing z1
                v[1][0] = v[1][0] * 0.7
                v[1][1] = v[1][1] * 0.3 
            for v in val_z2:
            # update belief according to observation
                #update for observing z2
                v[1][0] = v[1][0] * 0.3
                v[1][1] = v[1][1] * 0.7
            # recombine vectors
            value = []
            for v1 in val_z1:
                for v2 in val_z2:
                    action = 'u3' if  not (v1[0] == v2[0]) else v1[0]
                    value.append([action,[ v1[1][0] + v2[1][0], v1[1][1] + v2[1][1] ]])
            del(val_z1)
            del(val_z2)
            # include prediction about state transitions
            val_z1 = copy.deepcopy(value)
            value = []
            for v1 in val_z1:
                a = v1[1][0]
                b = v1[1][1]
                # new = [v1[0],[ -0.6*a + 0.6*b + 0.8*a + 0.2*b , 0.8*a + 0.2*b ] ]
                new = [v1[0],[ 0.2*a + 0.8*b , 0.8*a + 0.2*b ] ]
                new = [new[0], [new[1][0] - 1, new[1][1] - 1]]
                value.append(new)
            # prune values not contributing to maximum
            if pruning > 0:
                if pruning > 1:
                    self.prune_values(value, True)
                else:
                    self.prune_values(value, False)
            # add terminal actions
            value.extend([['u1',[-100, 100]], ['u2', [100, -50]] ])

        self.plot(value)
        print(value)
        print(len(value))
        self.value_vectors = value
        return value
        
    def get_reward_prediction(self, belief):
        # get best action
        best_reward = None
        best_action = None
        for v in self.value_vectors:
            expected = v[1][0]*belief + v[1][1]*(1 - belief) 
            if best_reward == None or expected > best_reward:
                best_action = v[0]
                best_reward = expected
        if DEBUG:
            print("best action, expected reward")
            print(best_action, best_reward)
        return best_reward

    def belief_update_sensing(self, true_state, belief):
        # obtain measurement
        measurement = 'z1' if random.random() < 0.7 else 'z2'
        if true_state == 'x2':
            measurement = 'z2' if measurement == 'z1' else 'z1'
        # belief update using Bayes rule
        if DEBUG:
            print("old belief:", belief)
            print("measurement:", measurement)
        if measurement == 'z1':
            p1 = belief # pravdepodpbnost x1
            bayes = (0.7 * p1 ) / ( (0.7 * p1) + (0.3 * (1 - p1)) ) # new p1
            belief =  bayes
        if measurement == 'z2':
            p2 = 1 - belief # pravdepodpbnost x2
            bayes = (0.7 * p2 ) / ( (0.7 * p2) + (0.3 * (1 - p2)) ) # new p2
            belief = 1 - bayes    
        return belief
    

    def simulate_outcome(self, horizon, start_belief = None):
        #parameters of the example:
            # p(z1|x1) = 0.7 p(z2|x1) = 0.3
            # p(z2|x1) = 0.3 p(z2|x2) = 0.7
        if start_belief != None:
            belief = start_belief
        else:
            belief = self.belief
        # reflect belief on real state
        true_state = 'x1' if random.random() < belief else 'x2'
        # true_state = random.choice(['x1', 'x2'])
        if DEBUG:
            print("True state: ",true_state)
        total_reward = 0
        best_action = None
        step = 0
        while not (best_action in self.terminal):
            #use sensing to update belief about state..
            belief = self.belief_update_sensing(true_state, belief)
            if DEBUG:
                print("new belief", belief)
            # get best action
            best_reward = None
            best_action = None
            for v in self.value_vectors:
                if step == horizon:
                    if v[0] == 'u3':
                        continue
                expected = v[1][0]*belief + v[1][1]*(1 - belief) 
                if best_reward == None or expected > best_reward:
                    best_action = v[0]
                    best_reward = expected
            step += 1
            if DEBUG:
                print("best action, expected reward")
                print(best_action, best_reward)
            total_reward += self.rewards[(true_state, best_action)]

            # measurements , state transitions, belief update...
            if best_action == 'u3':
                # state transition? 0.8 prop...need belief update...
                if random.random() < self.state_change_prop[best_action]:
                    true_state = 'x1' if true_state == 'x2' else 'x2' # change state
                if DEBUG:
                    print("new state:", true_state)
                #belief update:0.8 - 0.6p1
                belief = 0.8 - 0.6*belief
            if DEBUG:
                print("belief after possible transition:", belief)
                print('Total reward:', total_reward)       
        return total_reward


    def prune_values(self, vector, agressive):
        if agressive:
            error = -0.000001
        else: 
            error = 0.0000001
        safe = set([])
        for i1 in range(len(vector)):
            for i2 in range(i1 + 1,len(vector)):
                # print(vector[i1])
                # print(vector[i2])
                v1 = vector[i1]
                v2 = vector[i2]
                b = -1
                if v1 != v2:
                    try:
                        b = (v2[1][1] - v1[1][1]) / ( v1[1][0]-v1[1][1]-v2[1][0]+v2[1][1] )
                    except:
                        b = -1
                if  b > 0 and b < 1:
                    # print("point of intersection:", b)
                    val_v1 = v1[1][0]*b + v1[1][1]*(1 - b)
                    val_v2 = v2[1][0]*b + v2[1][1]*(1 - b)
                    # find if greater vector exists on [0,1]
                    v1_safe = True
                    v2_safe = True
                    for v3 in vector:
                        if agressive:
                            if v3 == vector[i1] or v3 == vector[i2]:
                                continue
                        val_v3 = v3[1][0]*b + v3[1][1]*(1 - b) #value at point b
                        if val_v3 > (val_v1 + error):
                            v1_safe = False
                            v2_safe = False
                    # if no vector has larger value in point b as v1 and v2, they must be parts of solution
                    if v1_safe:
                        safe.add((v1[1][0], v1[1][1])) # list to tuple
                    if v2_safe:
                        safe.add((v2[1][0], v2[1][1]))
       
        # now prune vectors not in safe set
        if DEBUG:
            print("unpruned len:", len(vector))
            print(safe)        
        i = 0
        while i < len(vector):
            tmp = (vector[i][1][0], vector[i][1][1]) # make tuple, ignore action
            if tmp not in safe:
                # print("PRUNING")
                vector.pop(i)
                i -= 1
            else:
                safe.discard(tmp)
            i += 1
        if DEBUG:
            print("pruned len:", len(vector))
            print(safe)
            print(vector)


    def plot(self, vec = None):
        # method to plot current value vectors over belief space
        if vec == None:
            vec = self.value_vectors
        plt.ylabel('expected reward')
        plt.xlabel('belief')
        plt.axis([0.0, 1.0, -100.0, 100.0])
        for v in vec:
            plt.plot([v[1][1], v[1][0]])
        plt.show()



if __name__ == "__main__" :
    print("this is Pomdp simulation...")
    starting_belief = 0.5
    p = Pomdp(starting_belief)
    N = 2000 # pocet iteracii simulacie
    horizon = 10
    vector = p.pomdp(horizon, 2) 
    print("belief    reward")
    for belief in range(0,101,2):
        belief /= 100.0
        reward = 0
        random.seed()
        for i in range(N):
            if DEBUG:
                print('-'*40)
            reward += p.simulate_outcome(horizon, belief)
        if DEBUG:
            pass
        else:
            # print belief,"  ", abs(p.get_reward_prediction(belief) - reward/N)
            print belief,"  ",  reward/N
            # print("Belief:", belief / 100.0)
            # print("Predicted reward:", p.get_reward_prediction(belief))
            # print('avg. reward:', reward/N)

    

   
